#!/usr/bin/ruby

# This script is designed to clean up old/unused revisions checked in
# the pclouds-vcsdeploy modules 'files' and 'templates' directories.
#
# 1.1 - Added support for regenerating the hash database during the maintenance run
# 1.2 - Changed files db to files_md5.yaml and removed separate 'templates' path
# 
# Install with puppet
# cron{ vcsdeploy-maintenance:
#   command => "/etc/puppet/environments/dev/modules/pclouds-vcsdeploy/bin/vcsdeploy-maintenance.rb 2592000",
#   user => 'root',
#   hour => 6,
#   minute => 15,
# }

# 	Copyright 2011 David McCormick
#  
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#  
#	http://www.apache.org/licenses/LICENSE-2.0
#  
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

require 'fileutils'
require 'find'
require 'digest/md5'
require 'time'
require 'syslog'
require 'yaml'
require 'pp'

def log(message)
  # $0 is the current script name
  Syslog.open($0.gsub(/^.*\//,''), Syslog::LOG_PID | Syslog::LOG_CONS) { |s| s.notice message }
end

def debug(message)
  # $0 is the current script name
  Syslog.open($0.gsub(/^.*\//,''), Syslog::LOG_PID | Syslog::LOG_CONS) { |s| s.info message } if DEBUG
end

# Set a default of 30days (2592000 seconds)
DEFAULT_MAXAGE=2592000
#DEFAULT_MAXAGE=30
DEBUG=true

start = Time.now

maxage = ARGV[0].to_f
maxage = DEFAULT_MAXAGE if (!maxage)
debug("The maxage is set to #{maxage} seconds") 

module_dir = File.dirname(__FILE__).gsub(/\/bin$/,'')
module_dir = Dir.pwd.gsub(/\/bin\/?$/,'') if (module_dir == '.')
file_dir = "#{module_dir}/files"
filehashes = "#{module_dir}/files_md5.yaml"
debug("The vcsdeploy module is installed in #{module_dir}")
log("Starting vcsdeploy-maintenance to purge unused files and templates in #{module_dir}")

# write the lock file
FileUtils.touch("#{module_dir}/maintenance.lock")

updatehashdb = false
if File.directory?(file_dir)
  Dir.new(file_dir).each {|entry|
    next if (entry =~ /^\.\.?$/)
    debug("#{file_dir} contains #{entry}")
    found_revision=false
    Dir.new("#{file_dir}/#{entry}").each {|revision|
      next if (revision =~ /^\.\.?$/)
      if File.directory?("#{file_dir}/#{entry}/#{revision}")
        debug("Found revision: #{revision}")
        found_revision=true
        mtime = File.mtime("#{file_dir}/#{entry}/#{revision}")
        elapsed = Time.now-mtime
        if (elapsed > maxage)
          FileUtils.rm_rf("#{file_dir}/#{entry}/#{revision}")
          log("deleted #{file_dir}/#{entry}/#{revision}: last used #{elapsed} seconds ago")
          updatehashdb = true
        else
          debug("keeping #{file_dir}/#{entry}/#{revision}: last used #{elapsed} seconds ago")
        end
      end
    }
    if (!found_revision)
      debug("Deleting empty #{file_dir}/#{entry}")
      FileUtils.rm_rf("#{file_dir}/#{entry}")
    end
  }
end

# now update the hash database if we deleted files...
# wait for an exclusive lock on the database file 
# to allow puppet runs checking out to finish

if (updatehashdb == true)
  debug("Waiting for exclusive lock on #{filehashes}")
  allfiles = {}
  if (File.exists?(filehashes))
    (file = File.new(filehashes,'r+')).flock(File::LOCK_EX)
    file.truncate(0)
    file.seek(0,IO::SEEK_SET)
  else
    (file = File.new(filehashes,'w')).flock(File::LOCK_EX)
  end

  debug("Regenerating database of md5 sums...")
  # add all of the files into the hash_database
  for dir in ['files','templates']
    debug("Calculating md5 sums of #{module_dir}/#{dir}")
    Find.find(file_dir) {|path|
      if File.file?(path)
        sum = Digest::MD5.file(path).to_s
        debug("#{sum} : #{path}")
        allfiles[sum] = path if (!allfiles[sum])
      end
    }
  end

  pp allfiles

  # write the database back out                                }
  debug("Writing and closing md5 database")
  file.write(allfiles.to_yaml.to_s)
  file.flock(File::LOCK_UN)
  file.close
end

# remove the maintenance lock again
FileUtils.rm("#{module_dir}/maintenance.lock") if File.file?("#{module_dir}/maintenance.lock")

ended=Time.now

log("vcsmaintenance finished in #{ended-start} seconds.")
