# This parser function checks out files from subversion and adds them to the puppet catalogue.
# 
# Version 0.29
# Dave McCormick
#
# Copyright 2012 David McCormick
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Change history: -
# 0.5  - added md5 sums of checked out files and hardlinking of duplicate files
# 			added interaction point for the maintenance script
# 0.6  - Added the vcsdeploy type and provider for local checkout and integrated with this function.
# 			The user will choose either 'local' or 'master' for 'checkout' (default is 'master')
# 0.7  - parser function becomes vcsdeploy and checkout => 'server' will cause it to
# 			add a vcsdeploy resource (with the current scope) to be processed on the server/agent.
# 0.8  - Add 'require' and 'notify' options
# 0.9  - Add 'templates' and 'dedupe' options
# 0.10 - Add a 'select' option where you can use a regex to choose a subset of the files
# 			from the checkout. This is only possible when the checkout is on the master, sorry.
# 0.11 - Prepare for the use of different providers by abstracting svn functions into
# 			svn_installed, svn_latest and svn_checkout functions.
# 0.12 - Add mechanism for the setting of user, group and mode explicitly per file or directory
# 0.13 - We need to purge files which don't belong within our checkout directories
# 0.14 - Fix notifyor caused by dirmode or filemode not being set.
# 0.15 - First steps to implementing git support: create svn_ and git_ methods which are called
#        according to which provider was selected. git_installed, git_validate and git_lookup implemented.  
# 0.16 - more git support work
# 0.17 - support selection or exclusion of files using special tags included in the optionial parameters block.
# 			filename.(tags=demo tomcat configs) = has three tags 'demo', 'tomcat' and 'configs'
# 0.18 - Another go at resolving my git support and shoe horning the concept of git branches, tags
#        and commit sha's into a single 'version' parameter, meanwhile dealing with the horribly limited
#        support for exporting specific revisions and sub branches from git repositories.
# 0.19 - Use a marshaled hash to store the md5 sums of files to improve execution performance.  Make sure
#        that only one process can copy the files and update the stored hash by using a flock exclusive lock.
# 0.20 - Problems with marshal so instead opting for yaml instead.  Fixed logic so that paths are the 
#        eventual paths and not the temporary ones.
# 0.21 - Write the file objects to a yaml file to elminate the 'find' to populate the file objects.
#        Extract locking to support methods
# 0.22 - Major overhaul, removing the separate 'files' and 'templates' path and using the
#        the yaml objects files to build the files lists to add to the puppet catalog.
#        Lots of cleaning up of complicated file paths in the code. 
#        Permission blocks now handled as separate methods.
# 0.23 - 'filemode' got lost in the move, fixed.
# 0.24 - add mode to the list of file_options.
# 0.25 - Fix notify regression
# 0.26 - Git not recongising requests for repos without an optional path
#        Changes to logging so that we work without stdlib
# 0.27 - Add a vcsdeployinfo object to each deploy to pass version information back to the client
#        nodes
# 0.28 - Put some more useful error checking around creating directories and make sure the 'temp'
#        directory has been created.
# 0.29 - Fix github issue 2 - 'require' option not allowed.  Also output made less verbose.
#        The stdlib fuctions, info and debug are assumed present.

# Requires: -
# svn (client) linux command 
# create_resources internal puppet function.
# ruby fileutils for rm -rf etc..
# ruby find
# md5 used to compare files saved on the master

require 'fileutils'
require 'pp'
require 'find'
require 'digest/md5'
require 'time'
require 'yaml'
require 'erb'

module Puppet::Parser::Functions

  # How long to wait for other processes checking out the files I want.
  MAX_WAIT_OTHER = 120
  MAX_WAIT_MAINTENANCE = 300

  # ------------------------------------
  # Usage notes: -
  # ------------------------------------

  # vcsdeploy[{deploy_opts},{defaults}] takes two hashes as arguments.
  # 'deploy_opts' is required and is a hash of the options we want to use for the deploy
  # 'defaults' is optional and the same, getting merged into 'deploy_opts' (where deploy_opts takes precedence).  
  # 'defaults' is useful for simplifying manifests as options can defined once in a varible and then passed to each call. 
  #
  # options (* required):-
  # puppet options: -
  #   *name - a unqiue name for the resource which can be used in 'path' and 'source' parameters. 
  #   *path - where to deploy to
  #   provider - (default svn) - only svn supported at present!
  #   checkout - 'server|agent' or 'master' (default is master) who performs the checkout?
  #   require - other puppet objects upon which this deploy requires
  #   notify - notify another puppet object when this object changes
  #   temp - where to checkout the files (default path_to_module/temp) 
  #   templates - turn off the template processing with (no|off|false|disable|disabled) 
  #   dedupe - turn off the file deduping on master with (no|off|false|disable|disabled) 
  # vcs options: -
  #   *source - where to check out from
  #   	svn: export from any directory within a repository
  #   		e.g. http://mysvnserver/repos/myrepo/a/directory/path
  #   	git: the git repository and optional path within it
  #   		e.g. file://path/to/repository.git/path/within/repository
  #   		or	 ssh://user@server:repository.git/path/within/repository
  #   		or	 git://server/repository.git
  #   		or   http[s]://user:password@server/path/to/repository.git/path/within/repository
  #   *version - the version (branch,tag,revision/commit#) we want to deploy
  #   		svn: revision (e.g. r378), tag/branch (e.g. testing), 'latest' - track the latest revision.
  #   		git: tree-ish: tag/branch (e.g. testing), commit hash (e.g. d9c13068)
  #   			** git ALWAYS tracks the latest unless a specific commit is provided **
  #   			** git tags are locked branches so WILL remain at the same version **
  #   user - user for the check out
  #   pass - password for the user
  #   select - only include files or directories which match this regex, which is applied to the full path.
  #    Where files or directories match but their parents do not match, they are 'bubbled' up the file
  #    hierarchy until they are placed in a directory that does exist or the root of the checkout.
  # file options: -
  #   Select/exclude files using 'tags' which must be set using a vsdeploy special block (see below)
  #     includetags - include only the files with the specified tags
  #     excludetags - exclude any files with the specified tags 
  #   owner - set the uid on the files and dirs
  #   group - set the group on the files and dirs
  #   dirmode - set this chmod mode on directories
  #   filemode - set this chmod mode on files
  #   selrange - What the SELinux range component of the context of the file should be
  #   selrole - What the SELinux role component of the context of the file should be
  #   seltype - What the SELinux type component of the context of the file should be
  #   seluser - What the SELinux type component of the context of the file should be
  #   replace - true|false|yes|no (default is true) - replace an existing file with different content.
  #   purge - true|false|yes|no (default is true) - remove files within the destination directories
  #    not explicity managed by puppet.  Files will not be deleted when removed from the vcs
  #    repository without this option.
  #
  #   The 'name' and 'version' parameters can both be inserted into the 'path' or 'source'
  #   parameters if we wish, using the following syntax: -   
  #   %%NAME%% - gets replaced by 'name'
  #   %%VERSION%% - gets replaced by 'version'
  #   e.g.
  #   source => 'http://svnserver/mypuppetrepo/modules/%%NAME%%/tags/%%VERSION%%',
  #   path => '/etc/httpd/%%NAME%%',
  #
  # Per object settings:  You can set owner, group, file permissions or tags on a per-file basis.
  #   vcsdeploy lets you do this by adding a special block to the filename, e.g.
  #   filename.(owner=puppet,group=puppet,mode=0755,tags=include config)
  #   The entries in the vcsdeploy control block change apply to only that object and not any others, and they are
  #   removed from the name when deployed.
  #   The arguments (in any order) are: -
  #     owner|user
  #     group
  #     mode
  #     selrange
  #     selrole
  #     seltype
  #     seluser
  #     tags

  newfunction(:vcsdeploy) do |args|
    # Make deep copies of the arguments - because we want to change them
    deploy_opts = Marshal::load(Marshal.dump(args[0]))
    default_opts = Marshal::load(Marshal.dump(args[1])) if (args[1])               

    debug("vcsdeploy called")
    start = Time.now

    # allow access to create_resources function...
    Puppet::Parser::Functions.autoloader.load(:create_resources) unless Puppet::Parser::Functions.autoloader.loaded?(:create_resources)

    # ------------------------------------
    # validate the parameters and setup
    # ------------------------------------

    # valid options
    valid_options = [ 'provider', 'path', 'owner', 'checkout', 'group', 'dirmode', 'filemode', 'source', 'user', 'pass', 'name', 'version', 'temp', 'replace', 'selrange', 'selrole', 'seltype', 'seluser', 'templates', 'dedupe', 'select', 'purge', 'includetags', 'excludetags', 'notify', 'require' ]
    required_options = [ 'name','path','source','version' ]
    type_options = [ 'provider', 'name', 'source', 'version', 'user', 'pass', 'path', 'owner', 'group', 'dirmode', 'filemod', 'selrange', 'selrole', 'seltype', 'seluser', 'scope', 'require', 'notify', 'templates' ]
    file_options = [ "owner", "group", "mode", "selrange", "selrole", "seltype", "seluser", "replace", "ensure" ]
    valid_providers = [ "svn", "git" ]

    # merge in the default_opts (if there are any), keep the deploy_opts value if there are duplicates.
    if (default_opts)
      deploy_opts.merge!(default_opts){|key, oldval, newval| oldval}
    end

    debug("Called with options:-")
    deploy_opts.each {|k,v|
      debug("#{k}: #{v}")
    }

    validate_options(deploy_opts,valid_options,required_options,type_options,file_options,valid_providers)

    # perform any provider specific validations
    send(:"#{deploy_opts['provider']}_validate",deploy_opts)

    # Allow selection of tagged objects using the the (tags=xyz abc) vcsnotice block in the filename.
    include_tags = []
    exclude_tags = []
    include_tags = deploy_opts['includetags'].split(' ') if deploy_opts['includetags']
    exclude_tags = deploy_opts['excludetags'].split(' ') if deploy_opts['excludetags']
    include_tags.each {|t| debug("include tag: #{t}") }
    exclude_tags.each {|t| debug("exclude tag: #{t}") }

    # replace any %%NAME%% tokens in 'path' or 'source'
    # we can't do %%VERSION%% until we know if we need to find the 'latest' version
    for option in ['path','source']
      deploy_opts[option].gsub!(/%%NAME%%/,deploy_opts['name']) if (deploy_opts['name'])
    end

    # ------------------------------------
    # return a 'Vcsdeploy' resource if checkout == 'server' or 'agent'
    # ------------------------------------

    # If checkout == server then create a vcsdeploy resource for the vcs type/provider to deploy
    # pass it all the parameters including the current puppet scope.
    if (deploy_opts['checkout'] =~ /^(server|agent)$/)
      if (deploy_opts['select'])
        notice "Sorry, you can only use the 'select' option when checking out on the puppetmaster."
        raise Puppet::ParseError, "Invalid use of 'select' option and agent checkout"
      end

      # Save the current scope as a hash to pass to the vcsdeploy resource.
      deploy_opts['scope'] = Marshal.dump(to_hash)

      # create the vcsdeploy object with only the options specific for the 
      # vcsdeploy type
      vcsdeploy_object = { } 
      vcsdeploy_object[deploy_opts['name']] = {} 
      type_options.each {|option|
        if ( deploy_opts[option] )
          #debug "setting #{option} to #{deploy_opts[option]}"
          vcsdeploy_object[deploy_opts['name']][option] = deploy_opts[option]
        end
      }
      #pp vcsdeploy_object

      # call the puppet 'create_resources' function to do this
      function_create_resources(['vcsdeploy',vcsdeploy_object])
      info "Added vcsdeploy object '#{deploy_opts['name']}' to the catalog"
      finish = Time.now 
      info "vcsdeploy call finished in #{finish - start} seconds."
      return
    end

    # make sure we have the vcs client software installed
    send(:"#{deploy_opts['provider']}_installed")

    # ------------------------------------
    # lookup proper version
    # ------------------------------------

    # if have specified 'latest' with subversion or a branch/tag with git
    # then we need to transform these into the actual revision which we need
    # to check out.
    send(:"#{deploy_opts['provider']}_lookup",deploy_opts)

    # now that we know the version, replace any %%VERSION%% tokens in 'path' or 'source'
    for option in ['path','source']
      deploy_opts[option].gsub!(/%%VERSION%/,deploy_opts['version']) if (deploy_opts['version'])
    end

    # ------------------------------------
    # Checkout will be processed on the master
    # ------------------------------------

    # lookup the location of this module
    deploy_opts['module_dir'] = File.dirname(__FILE__).gsub(/\/lib\/puppet\/parser\/functions$/,'')
    debug("The vcsdeploy module is located in #{deploy_opts['module_dir']}")
    deploy_opts['module_name'] = deploy_opts['module_dir'].gsub(/^.*\//,'')
    debug("The vcsdeploy module is called #{deploy_opts['module_name']}")

    # define our temporary directory if not done so by user...
    deploy_opts['temp'] = "#{deploy_opts['module_dir']}/temp" if (!deploy_opts['temp'])
    makedir(deploy_opts['temp']) if File.directory?(deploy_opts['temp'])
    # squash the 'source' path into a single directory name
    deploy_opts['source_squash']="#{deploy_opts['provider']}_#{deploy_opts['source'].gsub(/^.*:\/\//,'').gsub(/[^0-9A-Za-z._\-]/, '_')}"

    # make sure that the files and temp directories exist (we may have deleted them to clean up)
    makedir("#{deploy_opts['module_dir']}/files") if File.directory?("#{deploy_opts['module_dir']}/files")

    # ------------------------------------
    # Maintenance and clean up
    # ------------------------------------

    # Because we keep multiple versions of repositories on the puppet master
    # in the most space inefficient way possible (because we need to allow them
    # to be accessed by the regular file and template functions) we need to be
    # disciplined about clean up the files which are no longer 'in play' (needed).
    #
    # Check for the maintenance script lock and wait for it to finish or abort
    wait_lock("#{deploy_opts['module_dir']}/maintenance.lock",MAX_WAIT_MAINTENANCE)

    # ------------------------------------
    # check out if required... 
    # ------------------------------------

    versionlock = "#{deploy_opts['temp']}/#{deploy_opts['source_squash']}-#{deploy_opts['version']}.lock"
    versiontemp = "#{deploy_opts['temp']}/#{deploy_opts['source_squash']}-#{deploy_opts['version']}_#{$$}"
    versionprocessed = "#{versiontemp}_processed"
    versionobjectsfile = "vcsdeploy.yaml"

    filehashes = "#{deploy_opts['module_dir']}/files_md5.yaml"
    versiondir = "#{deploy_opts['module_dir']}/files/#{deploy_opts['source_squash']}/#{deploy_opts['version']}"

    # we must check if the requested revision is already available or check it out...
    # note a checkout could be comprised of only files or templates so we must check for EITHER of them!
    if (!(File.directory?(versiondir))) 	
      info("Checking out version #{deploy_opts['version']}")
      # check if this is already being checked out... and just wait for it to finish if it is.
      obtain_lock(versionlock)
      if (!(File.directory?(versiondir))) 	
        # We have the green light - it is OUR job to check out the files
        # Quick! Create the lock file so that other processes do not also try to checkout the files
        # (although we will be careful in case two or more processes reached here at the same time!)
        # call the provider specific checkout function/method
        begin
          send(:"#{deploy_opts['provider']}_checkout",deploy_opts,versiontemp)
        rescue
          FileUtils.rm_rf(versiontemp)
          release_lock(versionlock)
          raise Puppet::ParseError, "Couldn't export version #{deploy_opts['version']} of #{deploy_opts['source']}"
        end

        # ------------------------------------
        # add the files and templates to the store on the master
        # de-duplicate files to save on disk space
        # ------------------------------------

        # before we copy our checked out files, load the md5 hashes of other files
        # so that we can hard link to duplicate files to save space

        allfiles = {}
        if (deploy_opts['dedupe'] !~ /^(no|off|false|disable|disabled)$/)
          if (File.exists?(filehashes))
            (file = File.new(filehashes,'r+')).flock(File::LOCK_EX)
            allfiles = YAML.load(file)
          else
            (file = File.new(filehashes,'w')).flock(File::LOCK_EX)
          end
        end

        # Find all of the checked out files and copy them to the files or templates directories
        # Where the contents of the files match any of our md5 hashes then link to the existing file rather than
        # creating a duplicate file.  NOTE: files with different names but same content will be linked.
        update_hashes = false
        checkout_hashes = {}
        puppet_objects = {}
        Find.find(versiontemp) {|path|
          # Work out the relative and newpath 
          # add per object attributes if present in filename
          relative_path = path.gsub(/^#{versiontemp}/,'')
          relative_path = '/' if (relative_path == "")
          debug("Found #{relative_path}")
          if (has_permblock?(relative_path))
            attributes = read_permblock(relative_path)
            debug("Found following per object attributes:-")
            attributes.each {|k,v|
              debug("attribute #{k} = #{v}")
            }
            relative_path = remove_permblock(relative_path)
            debug("relative path changed to #{relative_path}")
            puppet_objects[relative_path] = attributes
          else
            puppet_objects[relative_path] = {}
          end
          newpath = "#{versionprocessed}#{relative_path}"
          debug("Processing #{path}")

          # If its a file, hardlink duplicate files.
          if File.file?(path)
            sum = Digest::MD5.file(path).to_s
            makedir(newpath.gsub(/[^\/]+$/,''))
            if deploy_opts['dedupe'] !~ /^(no|off|false|disable|disabled)$/
              if (checkout_hashes[sum] && File.exists?(checkout_hashes[sum]))
                # link to a file within the same checkout
                debug("Linking #{newpath} to local file #{checkout_hashes[sum]}")
                File.link(checkout_hashes[sum],newpath)
              elsif (allfiles[sum] && File.exists?(allfiles[sum])) 
                # link to the existing file
                debug("Linking #{newpath} to file #{allfiles[sum]}")
                File.link(allfiles[sum],newpath)
              else
                # copy the file
                notice("ERROR: #{allfiles[sum]} found in md5 db but does not exist!") if (allfiles[sum] && !File.exists?(allfiles[sum]))
                debug("Copying to #{newpath}")
                FileUtils.cp(path,newpath,:preserve => true)
                checkout_hashes[sum] = newpath
                #debug "Saving hash #{sum} : #{versiondir}#{relative_path}"
                allfiles[sum] = "#{versiondir}#{relative_path}"
                update_hashes = true
              end
            else
              debug("Creating #{newpath}")
              FileUtils.cp(path,newpath,:preserve => true)
            end
            puppet_objects[relative_path]['ensure'] = 'file'
          elsif File.directory?(path)
            puppet_objects[relative_path]['ensure'] = 'directory'
          end
          #debug("puppet objects:-")
          #puppet_objects.each {|k,v|
          #  debug "object: #{k}"
          #  v.each {|x,y| debug("#{x} = #{y}") }
          #}
        }
        checkout_hashes = {}

        # write the puppet objects yaml file.
        debug("Writing the puppet objects file: #{versionprocessed}/#{versionobjectsfile}")
        File.open("#{versionprocessed}/#{versionobjectsfile}",'w') {|f|
          YAML::dump(puppet_objects,f)
        }

        # remove the temporary files now that they have been processed and placed into templates and files...
        FileUtils.rm_rf(versiontemp)

        # make my copies live by renaming them, or get rid of them and wait if my checkout has collided with someone elses...
        if (File.directory?(versionprocessed))
          begin
            debug("making #{versiondir}")
            makedir(versiondir.gsub(/\/[^\/]+$/,''))
            debug("rename #{versionprocessed} to #{versiondir}")
            File.rename(versionprocessed,versiondir)
          rescue
            warn("Collision on creating #{versiondir}!") 
            warn("Removing #{versionprocessed}")
            FileUtils.rm_rf(versionprocessed)
            update_hashes = false
          end
        else
          raise Puppet::ParseError, "ERROR! I lost my processed files! : #{versionprocessed}"
        end

        # check if we need to add new hashes back into the hash file
        # read file again under exclusive lock so we don't miss any
        # in case it had been updated since we last read it.
        if (deploy_opts['dedupe'] !~ /^(no|off|false|disable|disabled)$/)
          if (update_hashes)
            debug("Writing updated md5 files hash database")
            file.truncate(0)
            file.seek(0,IO::SEEK_SET)
            file.write(allfiles.to_yaml.to_s) 
          end
          file.flock(File::LOCK_UN)
          file.close
        end
        allfiles = {}

        # success! - remove the lock
        release_lock(versionlock)
        debug("Files ready for distribution.")
      end

    else
      info("version #{deploy_opts['version']} is already available on the puppetmaster.")
    end

    # we now know that the files we need are indeed available

    # ------------------------------------
    # mark the version directory to show access
    # ------------------------------------
    # rather than writing a timestamp file, I'm going to 'touch' the directories instead
    # I can lookup the last time they were touched by looking at their mtime
    # e.g. File.mtime("xyz")
    # I'm using mtime rather than atime as atime updates are often disabled.

    if File.directory?(versiondir)
      FileUtils.touch(versiondir)
    end

    # ------------------------------------
    # Load the files into the catalog
    # ------------------------------------

    # load the default file options into their own hash for passing to create_resources
    # Add in any options such as 'user' or 'group' etc...
    filedefaults = {}
    file_options.each {|option|
      if ( deploy_opts[option] )
        filedefaults[option] = deploy_opts[option]
      end
    }

    possibleobjects = {}
    if (File.exists?("#{versiondir}/#{versionobjectsfile}"))
      File.open("#{versiondir}/#{versionobjectsfile}",'r') {|po|
        possibleobjects = YAML.load(po)
      }
    else
      raise Puppet::ParseError, "ERROR! I can't find my objects file! #{versiondir}/#{versionobjectsfile}"
    end
    #debug("possible objects:-")
    #possibleobjects.each {|k,v|
    #  debug "object: #{k}"
    #  v.each {|x,y| debug("#{x} = #{y}") }
    #}
    fileobjects = {}

    # if the user has set a select regex then use that to filter the results - otherwise use /.*/
    if (deploy_opts['select'])
      info "Filtering the checkout using the select filter: #{deploy_opts['select']}"
      match_regex = Regexp.new("#{deploy_opts['select']}")
    else
      match_regex = Regexp.new(".*")
      deploy_opts['select']=".*"
    end

    possibleobjects.each {|path,attributes|
      srcpath = "#{versiondir}#{path}"
      destpath = "#{deploy_opts['path']}#{path}"
      debug "relative: #{path}"
      debug "source: #{srcpath}"
      debug "destination: #{destpath}"

      include_object = false
      # does the relative path match our seletion regex?
      # We always need to select the root path /
      if (path == "/" || path =~ match_regex)
        include_object = true
        # if includetags were specified check that the file has one of them...
        if (path != "/" && !include_tags.empty?)
          if (attributes['tags'] && !(include_tags & (attributes['tags'].split(' '))).empty?)
            debug("Accepted: #{path} by includetags")
          else
            include_object = false
            debug "Rejected: #{path} by includetags"
          end
        end
        # if excludetags was specified then check that this object doesn't match any of them...
        if (path != "/" && !exclude_tags.empty?)
          if (attributes['tags'] && !(exclude_tags & (attributes['tags'].split(' '))).empty?)
            include_object = false
            debug "Rejected: #{path} by excludetags"
          else
            debug "Accepted: #{path} by excludetags"
          end
        end
        if (include_object)
          debug "Selected: #{path}"
          # take all file attributes
          fileobjects[destpath] = {}
          attributes.each {|att,val|
            if (file_options.include?(att))
              #debug("Setting #{att} : #{val}")
              fileobjects[destpath][att] = val
            end
          }
          # add 'notify' to the object if rerquested...
          fileobjects[destpath]['notify'] = deploy_opts['notify'] if (deploy_opts['notify'])
          if (fileobjects[destpath]['ensure'] == 'directory')
            # set the purge attribute
            # without this files which are no longer present in the vcs repository will
            # continue to remain on the servers.  BE VERY CAREFULL IF USING SELECT! [scared!]
            if (deploy_opts['purge'] =~ /^(on|yes|true)$/)
              fileobjects[destpath]['purge'] = 'true'
              fileobjects[destpath]['recurse'] = 'true'
            end
            # set the mode explicitly or if there's no per object mode and a 'dirmode' option
            if (!fileobjects[destpath]['mode'] && deploy_opts['dirmode'])
              fileobjects[destpath]['mode'] = deploy_opts['dirmode']
            end
            # if this is the top level directory then add 'require' to the objects (if provided)...
            if (path == "" && deploy_opts['require'])
              debug "Adding require #{deploy_opts['require']} on the base path"
              fileobjects[destpath]['require'] = deploy_opts['require']
            end
          elsif (fileobjects[destpath]['ensure'] == 'file')
            # set the mode explicitly or if there's no per object mode and a 'filemode' option
            if (!fileobjects[destpath]['mode'] && deploy_opts['filemode'])
              fileobjects[destpath]['mode'] = deploy_opts['filemode']
            end
            if (destpath =~ /\.erb$/)
              # it's a template!  Use puppets template function to proivide the files contents.
              debug("template: #{destpath}")
              oldobject = fileobjects[destpath]
              fileobjects.delete(destpath)
              fileobjects[destpath.gsub(/\.erb$/,'')] = oldobject
              # REMOVED - Rendering the template manually using erb in favour of using puppets template function with a full path.
              #thetemplate = File.open(srcpath,'r') {|f| f.read }
              #renderer = ERB.new(thetemplate)
              #content = renderer.result(binding)
              #debug("rendered ok")
              #fileobjects[destpath.gsub(/\.erb$/,'')]['content'] = content
              fileobjects[destpath.gsub(/\.erb$/,'')]['content'] = function_template(["#{deploy_opts['module_dir']}/files/#{deploy_opts['source_squash']}/#{deploy_opts['version']}#{path}"])
            else
              # it's a file!  We'll serve this as a regular file.
              debug("file: #{destpath}")
              fileobjects[destpath]['source'] = "puppet:///modules/#{deploy_opts['module_name']}/#{deploy_opts['source_squash']}/#{deploy_opts['version']}#{path}"
            end
          end
        else
          debug "Skipped: #{path}"
        end
      else
        debug "Skipped: #{path} does not match #{deploy_opts['select']}"
      end
    }

    # We have our file objects
    # If we applied a select filter then we're going to need to process each path
    # and remove directories which do not exist, bubbling everything up into
    # directories which do exist or up to 'path'
    if (deploy_opts['select']) 
      debug "Make sure we have a usable object tree after the select filter has been applied"
      # find all paths and sort them in descending order
      # order is important, we need to change the paths from the bottom up
      fileobjects.keys.sort{|a,b| b <=> a}.each {|object|
        relativepath = object.gsub(/^#{deploy_opts['path']}\/*/,'')
        newpath = deploy_opts['path'].clone
        parsed = deploy_opts['path'].clone
        relativepath.split(/\/+/).each {|part|
          parsed << "/" << part
          newpath << "/" << part if fileobjects[parsed]	
        }
        if (newpath != object)
          debug "Renamed #{object} to #{newpath}"
          fileobjects[newpath]=fileobjects[object]	
          fileobjects.delete(object)
        end
      }
    end

    #debug("File objects:-")
    #fileobjects.each {|k,v|
    #  debug "object: #{k}"
    #  v.each {|x,y| debug("#{x} = #{y}") }
    #}

    # now add the resultant fileobjects in the catalog...
    # call the puppet 'create_resources' function to do this
    debug("Creating resources...")
    function_create_resources(['file',fileobjects,filedefaults])
    debug "Added #{fileobjects.keys.length} 'File' objects to the catalog"

    # Add a vcsdeployinfo resource
    vcsinfo = { 
		 deploy_opts['path'] => {
        :source => deploy_opts['source'],
        :version => deploy_opts['version']
       }
    }
    function_create_resources(['vcsdeployinfo', vcsinfo])

    # ------------------------------------
    # Finished!
    # ------------------------------------

    finish = Time.now 
    info "vcsdeploy call finished in #{finish - start} seconds."

  end

end

# ------------------------------------
# Support Methods
# ------------------------------------

def makedir(directory) 
	begin
		FileUtils.mkdir_p(directory)
	rescue
		notice "I could not create directory \"#{directory}\", please check permissions!"
		notice "The vcsdeploy module directories need to be writable by the user which you are running your puppet master as."
		raise Puppet::ParseError, "I don't have permission to create the directory #{directory}"
	end
end


def wait_lock(lock,maxwait = 120)
  if (File.file?(lock))
    notice("The lock file #{lock} exists so will wait")
    waitcount = 0
    while (File.file?(lock))
      sleep 5
      waitcount += 5
      info("Waited #{waitcount} of #{maxwait}")
      if (waitcount >= maxwait) 
        notice "Timing out waiting for the lock #{lock}"
        notice "Sorry, I can't continue or wait any longer."
        notice "Please run again later and if the error still persists then manual intervention may be required to remove the blockage!"
        raise Puppet::ParseError, "Timed Out waiting for #{lock}"
      end 
    end
  else
    debug "Lockfile #{lock} does not exist, proceeding"
  end
end

def obtain_lock(lock,maxwait = 120)
  wait_lock(lock)
  FileUtils.touch(lock)
end

def release_lock(lock)
  FileUtils.rm_f(lock) if File.file?(lock)
end

# ------------------------------------
# validate the parameters and setup
# ------------------------------------

def validate_options(options,valid_options,required_options,type_options,file_options,valid_providers)
  options.each_pair {|option, value|
    if !valid_options.include?(option)
      notice "Sorry! #{option} is not a recognised option of vcsdeploy!"
      raise Puppet::ParseError, "Invalid option #{option} when calling vcsdeploy"
    end
  }

  # validate the providers (or select the default, svn)
  if (options['provider'])
    if !valid_providers.include? options['provider']
      notice "Sorry! #{options['provider']} is not a recognised provider.  The supported providers are #{valid_providers.join(', ')}"
      raise Puppet::ParseError, "Invalid vcsdeploy provider #{options['provider']}"
    end
  else
    info "No provider specified - defaulting to 'svn'"
    options['provider'] = 'svn'
  end

  # check the required options have been specified.
  required_options.each {|option|
    if (!options[option])
      notice "Missing required option #{option}!"
      raise Puppet::ParseError, "Sorry, you must specify the #{option} when calling vcsdeploy"
    end
  }

  # checkout should be 'agent' or 'master' - default to 'master'
  if (options['checkout'])
    if (options['checkout'] !~ /^(agent|server|master)$/)
      notice "Sorry! 'checkout' needs to be 'server', 'agent' or 'master'"
      raise Puppet::ParseError, "Invalid option checkout => #{options['checkout']} when calling vcsdeploy"
    end
  else
    options['checkout'] = 'master'
  end

  # set a default purge option of 'true'
  if (options['purge'])
    if (options['purge'] !~ /^(true|false|yes|no|on|off)$/)
      notice "Sorry! 'purge' needs to be (true|false|yes|no|on|off)"
      raise Puppet::ParseError, "Invalid option purge => #{options['purge']} when calling vcsdeploy"
    end
  else
    options['purge'] = 'true'
  end
end

# -------------------------------------------
# Functions that deal with permissions blocks
# -------------------------------------------

def has_permblock?(path)
  return (path =~ /\.\(((,?(user|owner|group|selrange|selrole|seltype|seluser|tags)=[a-zA-Z0-9 ]+)|(,?mode=([0-7]+|[ugoa=+,rwxtsX-]+)))+\)$/)
end

def remove_permblock(path)
  return path.gsub(/\.\(.+\)$/,'')
end

def read_permblock(path)
  # handle files or directories with explicit permissions defined using a special permissions block at
  # the end of the filename.
  # e.g. filename.(owner=tomcat,group=tomcat,mode=755)
  # The permission block MUST be at the end of the file or directory name!

  lexer = { 'owner'    => /(user|owner)=([0-9]+|[a-zA-Z]+)/,
    'group'    => /group=([0-9]+|[a-zA-Z]+)/, 
    'mode'     => /mode=([0-7]+|[ugoa=+,rwxtsX-]+)/,
    'selrange' => /selrange=[a-zA-Z0-9]+/,
    'selrole'  => /selrole=[a-zA-Z0-9]+/,
    'seltype'  => /seltype=[a-zA-Z0-9]+/,
    'seluser'  => /seluser=[a-zA-Z0-9]+/,
    'tags'     => /tags=[a-zA-Z0-9 ]+/ 
  }

  attributes = {}
  if (has_permblock?(path))
    perobject_permissions = path.match(/\.\(((,?(user|owner|group|selrange|selrole|seltype|seluser|tags)=[a-zA-Z0-9 ]+)|(,?mode=([0-7]+|[ugoa=+,rwxtsX-]+)))+\)$/).to_s
    debug "perobject_permissions=#{perobject_permissions}"
    lexer.each {|att,rgx| 
      value = perobject_permissions.match(rgx).to_s.gsub(/^[^=]+=/,'') 
      if (value && value != "")
        #debug "Setting perobject_#{att} to #{value} on #{path}" 
        attributes[att] = value 
      end
    }
  else
    debug("No perobject permissions on #{path}")
  end
  return attributes
end

# ------------------------------------
# svn / subversion support
# ------------------------------------

def svn_installed
  # make sure we have the 'svn' client installed.
  svn_avail=`which svn 2>&1`
  if ( svn_avail =~ /no svn in/)
    notice("I can't lookup the latest version because I can't find the svn command!")
    raise Puppet::ParseError, "svn command not found on puppetmaster"
  end
  return true
end

def svn_validate(deploy_opts)
  # lowercase any svn revisions begining with 'R' instead of 'r'
  if ( deploy_opts['provider'] == 'svn' )
    if ( deploy_opts['version'] =~ /^R([0-9]+)$/)
      deploy_opts['version'] = "r#{$1}"
    end
  end
end

def svn_lookup(deploy_opts)

  # if we have requested the 'latest' version then we'd better go and lookup which that is...
  if (deploy_opts['version'] == 'latest')
    # NOTE: there in no use putting a %%VERSION%% token in the source if you have selected
    # the version 'latest'!  It has no way of looking up the version!
    if (deploy_opts['source'] =~ /%%VERSION%%/)
      notice "Sorry, you CAN'T use the %%VERSION%% token with the version 'latest'"
      raise Puppet::ParseError, "Impossible source and version combination supplied.  I can't work out the source or the version number because they depend on each other!"
    end

    info("vcsdeploy: Deploying latest version of #{deploy_opts['source']}")
    infocmd = "svn info #{deploy_opts['source']} --no-auth-cache --non-interactive"
    if ( deploy_opts['user'] != '' && deploy_opts['pass'] != '' )
      # add authentication information to the command line
      infocmd << " --username #{deploy_opts['user']} --password #{deploy_opts['pass']}"
    end

    debug("Running svn command: #{infocmd.gsub(/--password\s+\S+/,'--password #####')}")
    output = `#{infocmd} 2>&1`
    # check it completed successfully?
    if ( $?.exitstatus != 0 )
      notice("Exit code was: #{$?.exitstatus}")
      notice("#{output}")
      raise Puppet::ParseError, "Can't lookup latest revision of #{deploy_opts['source']}"
    else
      output =~ /Last Changed Rev:\s*([0-9]+)/m
      if ( $1 == "" )
        notice("The last changed revision is empty!")
        raise Puppet::ParseError, "Can't lookup latest revision of #{deploy_opts['source']}"
      else
        # set the version to version to what we just looked up.
        deploy_opts['version'] = "r#{$1}"
        info("The latest version is : #{deploy_opts['version']}")
      end
    end
  else
    info("vcsdeploy: Deploying version #{deploy_opts['version']} of #{deploy_opts['source']}")
  end
end

def svn_checkout(deploy_opts,tempdir)
  # check out to our own directory in case we are still colliding with other processes
  checkoutcmd = "svn export #{deploy_opts['source']} #{tempdir} --no-auth-cache --non-interactive"

  if ( deploy_opts['version'] =~ /^r([0-9]+)$/)
    # we've been asked to install a specific svn revision.
    svnrevision = $1
    checkoutcmd << " --revision #{svnrevision}"
  end

  if ( deploy_opts['user'] != '' && deploy_opts['pass'] != '' )
    # add authentication information to the command line
    checkoutcmd << " --username #{deploy_opts['user']} --password #{deploy_opts['pass']}"
  end

  # check out the new version
  debug("Running svn command: #{checkoutcmd.gsub(/--password\s+\S+/,'--password #####')}")
  output = `#{checkoutcmd} 2>&1`
  # check it completed successfully?
  if ( $?.exitstatus != 0 )
    notice("Exit code was: #{$?.exitstatus}")
    notice("#{output}")
    raise "Svn export failed."
  else
    debug "Files checked out ok."
  end
  return true
end

# ------------------------------------
# git support
# ------------------------------------

def git_installed
  # make sure we have the 'git' client installed.
  git_avail=`which git 2>&1`
  if ( git_avail =~ /no git in/)
    notice("I can't lookup the latest version because I can't find the git command!")
    raise Puppet::ParseError, "git command not found on puppetmaster"
  end
  return true
end

def git_validate(deploy_opts)
  # work out the git options and detect issues with the arguments
  if (deploy_opts['source'] !~ /\.git(\/.*)?$/ )
    notice("Sorry, a source git repository MUST include a .git in the source parameter!")
    raise Puppet::ParseError, "git sources MUST include the .git so I can work out the repository location!"
  else
    # Any path after the .git is a sub path - add specific options for repository and subpath
    deploy_opts['gitsubpath'] = deploy_opts['source'].gsub(/^.*\.git/,'')
    deploy_opts['gitrepository'] = deploy_opts['source'].gsub(/\.git(\/.*)?$/,'.git')

    # depending on the type of the transport protocol, we may need to add username and password
    # validate acceptable use of these
    if (deploy_opts['user'])
      if (deploy_opts['password'])
        # only allow a password parameter when using http/https
        if (deploy_opts['gitrepository'] =~ /https?:\/\//)
          deploy_opts['gitrepository'] = deploy_opts['gitrepository'].gsub(/^(https?:\/\/)/,"\\1#{deploy_opts['user']}:#{deploy_opts['password']}@")
        else
          raise Puppet::ParseError, "You can only specify a user and password with the http/https git protocol"
        end
      else
        # Only allow a username with http/s and ssh transports
        if (deploy_opts['gitrepository'] =~ /ssh:\/\//)
          deploy_opts['gitrepository'] = deploy_opts['gitrepository'].gsub(/^ssh:\/\//,"ssh://#{deploy_opts['user']}@")
        else
          raise Puppet::ParseError, "You can only specify a passwordless user with the ssh git protocol"
        end
      end
    end
  end
end

def git_lookup(deploy_opts)

  # --------------------------------------------------------------------------------
  # handle a branch or tag and find the latest commit sha
  # --------------------------------------------------------------------------------
  # The version can be the name of a tag, branch, or commit sha or branch:commitsha
  # Branches and tags without :commit we'll lookup the latest sha
  # Sha's without a branch: or tag: on the front of them will be assumed to be the master
  # branch.  At the end of this function the version should be of the form branch:sha

  if (deploy_opts['version'] =~ /^[a-zA-Z0-9]+:[a-f0-9]+/)
    info("The version has been been specified as a branch/commit by user, no lookup required")
  else
    # look up the tag or branchs for latest commit shas
    if (deploy_opts['version'] == 'latest')
		info("Assuming that the 'latest' version should be the latest commit on the branch 'master'")
    	deploy_opts['version'] = 'master'
    end
    lookupcmd = "git ls-remote #{deploy_opts['gitrepository']} #{deploy_opts['version']}" 
    debug("Running git command: #{lookupcmd}")
	 result = `#{lookupcmd} 2>&1`
    # check it completed successfully?
    if ( $?.exitstatus != 0 )
      notice("ERROR: Exit code was: #{$?.exitstatus}")
      notice("ERROR: Result: #{result}")
      raise Puppet::ParseError, "Can't lookup branches/tags in #{deploy_opts['gitrepository']}"
    else
      commitsha = result.gsub(/\s.*$/,'')
      if (commitsha != "")
        # update the version with the actual commit hash e.g. branch1:e29023baa7f9e506f0113fe136a09fd42aa90774
        info("The latest version of #{deploy_opts['version']} is #{commitsha}")
        deploy_opts['version'] = "#{deploy_opts['version']}:#{commitsha}"
      else
        info("#{deploy_opts['version']} isn't a branch or a tag so assuming that it is a specific commit hash from the master branch")
        deploy_opts['version'] = "master:#{deploy_opts['version']}"
      end
    end 
  end
end

def git_checkout(deploy_opts,tempdir)
  # Try to check out the git branch from remote source only bringing down the minimum of files/history.
  # This is presently far less effecient than subversion in that we need to fetch an entire branch. 

  # check out to our own directory in case we are still colliding with other processes
  branch = deploy_opts['version'].gsub(/:.*$/,'')
  commit = deploy_opts['version'].gsub(/^.+:/,'')

  # Fetch the remote branch locally...
  makedir(tempdir)
  Dir.chdir(tempdir)
  result = `git init #{tempdir} 2>&1`
  if ($?.exitstatus != 0)
    notice("Couldn't create empty git repository, #{tempdir}");
    notice(result)
    raise "Couldn't create empty git repository, #{tempdir}"
  end
  result = `git fetch #{deploy_opts['gitrepository']} #{branch} 2>&1`  	
  if ($?.exitstatus != 0)
    notice("Couldn't fetch the branch #{branch} from the remote repository #{deploy_opts['gitrepository']}");
    notice(result)
    raise "Couldn't fetch the branch #{branch} from the remote repository #{deploy_opts['gitrepository']}"
  end

  # use git archive to extract the files (using the subpath if defined)
  if (deploy_opts['gitsubpath'])
    result = `git archive #{commit} #{deploy_opts['gitsubpath'].gsub(/^\//,'')} | tar -x`
  else
    result = `git archive #{commit} | tar -x`
  end
  if ($?.exitstatus != 0)
    notice("Couldn't checkout commit #{commit} from the fetched branch!");
    raise "Couldn't checkout commit #{commit} from the fetched branch!"
  end

  #remove the .git directory
  FileUtils.rm_r('.git')

  return true
end

