
# Version 0.11

# Copyright 2012 David McCormick
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Change history
# v0.6 - This existed as ig-puppetmodule previously and has been modified to become part of vcsdeploy. Clean up. Make work.
# v0.7 and 0.8 - more work on passing the scope to vcsdeploy object for use with templates.
# v0.9 - add templates option
# v0.10 - allow per file owner, group and mode setting
# v0.11 - we were not setting owners, groups or modes

require 'fileutils'
require 'pp'
require 'erb'
require 'facter'

Puppet::Type.type(:vcsdeploy).provide(:svn) do
  desc "SVN Provider for deploying from subversion or git etc.."

  # Only allow the provider if svn is installed.
  commands :svncmd => 'svn'

  def create 
    notice "Checking out version #{@resource_copy[:version]} of #{@resource_copy[:source]} to #{@resource_copy[:path]}"

    checkoutcmd = [ "export", @resource_copy[:source], @resource_copy[:path], "--no-auth-cache", "--non-interactive" ]
    if ( @resource_copy[:version] =~ /^[rR]([0-9]+)$/)
      # we've been asked to install a specific svn revision.
      svnrevision = $1
      #print "Checking out svn revision #{svnrevision}\n"
      checkoutcmd << "--revision" << svnrevision
    end

    if ( resource[:user] != '' && resource[:pass] != '' )
      # add authentication information to the command line
      checkoutcmd << "--username" << "#{resource[:user]}" << "--password" << "#{resource[:pass]}"
    end

    # run the svn export command! 
    output = svncmd *checkoutcmd
    # check it completed successfully
    if ( $?.exitstatus != 0 )
      err "Exit code was: #{$?.exitstatus}\n"
      err "Sorry!, I couldn't export the requested files.\n"
      err "#{output}\n"
      raise ExecutionError "Couldn't export the resource"
    else
      # record the version number	
      if File.directory? "#{@resource_copy[:path]}"
        open("#{@resource_copy[:path]}/.version", 'w') { |f|
          f.puts @resource_copy[:version]
        }
      else
        err "Error, sorry #{resource[:name]} did not download properly."
        raise ExecutionError "Sorry #{resource[:name]} did not download properly."
      end		

      # process any templates before doing any explicit permissions
      # the horrible looking regex is to allow the explicit permission block
      # on the end of the name
      if (@resource_copy[:templates] !~ /^(no|false|off|disable|disabled)$/)
        # Now go and process any templates which have been deployed
        Find.find(@resource_copy[:path]) {|path|
          if (File.file?(path) && path =~ /\.erb(\.\(((,?(user|owner|group)=([0-9]+|[a-zA-Z]+))|(,?mode=[0-7]+))+\))?$/)
            debug "Processing template: #{path}"
            # Read and create the template
            thetemplate = File.open(path,'r') {|f| f.read }							
            renderer = ERB.new(thetemplate)
            # Write the processed template
            File.open(path.gsub(/\.erb(\.\(((,?(user|owner|group)=([0-9]+|[a-zA-Z]+))|(,?mode=[0-7]+))+\))?$/,'\1'),'w') {|f| f.write(renderer.result(binding)) }
            FileUtils.rm(path)
          end
        }
      else
        info "Templates have been disabled."
      end

      # process any files or directories with explict permissions
      # look for .(owner=xyz,group=xyz,mode=123) at the end of the name
      # modes must be the octal representations only!
      Find.find(@resource_copy[:path]) {|path|
        explicit_owner = ""
        explicit_group = ""
        explicit_mode = ""
        if (path =~ /\.\(((,?(user|owner|group)=([0-9]+|[a-zA-Z]+))|(,?mode=[0-7]+))+\)$/)
          # Remove the permissions block from the destination file name
          newpath = path.gsub(/\.\(((,?(user|owner|group)=([0-9]+|[a-zA-Z]+))|(,?mode=[0-7]+))+\)$/,'')
          # rename the file
          FileUtils.mv(path,newpath)
          explicit_permissions = path.match(/\.\(((,?(user|owner|group)=([0-9]+|[a-zA-Z]+))|(,?mode=([0-7]+|[ugoa=+,rwxtsX-]+)))+\)$/).to_s
          explicit_owner = explicit_permissions.match(/(user|owner)=([0-9]+|[a-zA-Z]+)/).to_s.gsub(/^(user|owner)=/,'')
          explicit_group = explicit_permissions.match(/group=([0-9]+|[a-zA-Z]+)/).to_s.gsub(/^group=/,'')
          explicit_mode = explicit_permissions.match(/mode=[0-7]+/).to_s.gsub(/^mode=/,'')
        else
          newpath = path
          explicit_owner = @resource_copy[:owner] if (@resource_copy[:owner])
          explicit_group = @resource_copy[:group] if (@resource_copy[:group])
          explicit_group = @resource_copy[:mode] if (@resource_copy[:mode])
        end
        if (explicit_owner != "")
          debug "Setting explicit owner of #{explicit_owner} on #{newpath}"
          FileUtils.chown(explicit_owner,nil,newpath)
        end
        if (explicit_group != "")
          debug "Setting explicit group of #{explicit_group} on #{newpath}"
          FileUtils.chown(nil,explicit_group,newpath)
        end
        if (explicit_mode != "")
          debug "Setting explicit mode of #{explicit_mode} on #{newpath}"
          FileUtils.chmod(explicit_mode.to_i(8),newpath)
        end
      }
    end
  end

  # remove the deploy if "ensure => absent" has been selected.
  def destroy
    notice "Removing #{@resource_copy[:path]}"
    FileUtils.rm_rf(@resource_copy[:path])
  end

  # does the deploy exist? Test if its directory is there.
  def exists?
    load_defaults
    debug "exists? #{resource[:name]}, path #{@resource_copy[:path]}"
    exists = File.directory?(@resource_copy[:path])
    debug "#{@resource_copy[:path]} exists?: #{exists}"
    exists
  end

  # check if the version installed matches the requested version.
  def version
    version = ""

    # it is very simple to check the version when not 'latest' - just read the .version file. 
    if (File.directory?(@resource_copy[:path]))
      if (File.exists?("#{@resource_copy[:path]}/.version"))
        # read the version from its version file.
        file = File.open("#{@resource_copy[:path]}/.version")	
        version = file.read.chomp
        file.close
      end
    else
      err "Error! Sorry, I couldn't find the version of #{resource[:name]} yet I know it exists!" 
      raise "Error! Sorry, I couldn't find the version of #{resource[:name]} yet I know it exists!" 
    end

    # if the original version was set to latest we need to return 'latest' if the local resource is at the latest
    # version, otherwise return the actual version number...
    info "Local #{@resource_copy[:path]} is at version #{version}"
    if (resource[:version] == 'latest' && @resource_copy[:version] == version )
      return 'latest'
    end
    version
  end

  # If the versions don't match, call the destroy and create to update to the latest version.
  def version=(newversion)	
    notice "Version of #{resource[:name]} does not match the required version."
    destroy
    create
  end

  # ---------------------------------------------------------------------------------------------------------------------------
  # Helper Functions....

  def load_defaults()
    debug "load_defaults..."

    # debugging...
    debug "resource path : #{@resource[:path]}"
    debug "resource source : #{@resource[:source]}"
    debug "resource version : #{@resource[:version]}"

    # make a copy of the resource being worked upon .. if there isn't already a copy...
    # use a simple hash, we don't need anything fancy
    valid_options = [ 'path', 'owner', 'group', 'dirmode', 'filemode', 'source', 'user', 'pass', 'name', 'version', 'selrange', 'selrole', 'seltype', 'seluser', 'templates' ]
    @resource_copy = {}

    debug "creating resource_copy for #{resource[:name]}"
    valid_options.each {|option|
      if (option && resource[option.to_sym])
        @resource_copy[option.to_sym] = resource[option.to_sym]
      end
    }

    # optionally load in a scope passed with the vcsdeploy resource
    if @resource['scope']
      myscope = Marshal.load(@resource['scope'])
      debug "Loading scope variables from puppet master..."
      #puts "My loaded scope is: "
      #pp myscope
      myscope.each do |k,v|
        instance_variable_set("@#{k}",v)
        #debug "Setting @#{k}."
      end
    end

    # We can't continue without a path or source.
    raise ArgumentError "You need to set the 'path' paramater when defining a vcsdeploy" if ( @resource_copy[:path] == '')
    raise ArgumentError "You need to set the 'source' paramater when defining a vcsdeploy" if ( @resource_copy[:source] == '')

    # replace optional %%NAME%% and %%VERSION%% placeholders in the 'source' string.
    @resource_copy[:path].gsub!(/%%NAME%%/,resource[:name])
    @resource_copy[:source].gsub!(/%%NAME%%/,resource[:name])

    if (@resource_copy[:version] == 'latest')
      # NOTE: there in no use putting a %%VERSION%% token in the source if you have selected
      # the version 'latest'!  It has no way of looking up the version!
      if (@resource_copy[:source] =~ /%%VERSION%%/)
        err "Sorry, you CAN'T use the %%VERSION%% token with the version 'latest'"
        raise Puppet::ParseError, "Impossible source and version combination supplied.  I can't work out the source or the version number because they depend on each other!"
      end

      infocmd = [ "info", @resource_copy[:source], "--no-auth-cache", "--non-interactive" ]
      if ( resource[:user] != '' && resource[:pass] != '' )
        # add authentication information to the command line
        infocmd << "--username" << "#{resource[:user]}" << "--password" << "#{resource[:pass]}"
      end

      info "Looking up 'latest' version of #{@resource_copy[:source]}"
      output = svncmd *infocmd
      # check it completed successfully?
      if ( $?.exitstatus != 0 )
        err "Exit code was: #{$?.exitstatus}\n"
        err "Sorry!, I couldn't find the svn info for the requested deploy.\n"
        err "#{output}\n"
        raise ExecutionError "Couldn't find requested deploy"
      else
        output =~ /Last Changed Rev:\s*([0-9]+)/m 
        if ( $1 == "" )
          raise ExecutionError "I couldn't work out the latest revision!"
        else
          latest = "r#{$1}"
        end
      end
      info "The 'latest' version is #{latest}"
      @resource_copy[:version] = latest
    end

    # now that we know the version for sure we can replace it in the source and path parameters
    @resource_copy[:source].gsub!(/%%VERSION%%/,@resource_copy[:version])
    @resource_copy[:path].gsub!(/%%VERSION%%/,@resource_copy[:version])
    debug "resulting path : #{@resource_copy[:path]}"
    debug "resulting source : #{@resource_copy[:source]}"
    debug "resulting version : #{@resource_copy[:version]}"
  end

end
